unit Utility;

interface

uses SysUtils, Classes, League, Games, Lists, WinSock, ZoomReport;

type
  TScoreboardStyle = (ssColumns, ssDetailed);

  TPrefs = record
    GridHigh, GridWide, GridPlayers: Integer;  // size of team box grid, and height of each box
    SortMode: Integer;        // 0 = sort by alias; 1 = sort by colour; 2 = sort for Lord of the Ring, i.e. sort by alias, but rank within each colour, so there's a first on red, a first on blue, etc.
    NexusScores: Boolean;     // If true, let Nexus calculate Nexus scores.  If false, we will recalculate them ourselves.
    NexusOneShotDenyFriend,
    NexusOneShotDenyEnemy,
    NexusTwoShotDenyFriend,
    NexusTwoShotDenyEnemy: Integer;  // The are the ng_event_types for Vector's version of the Nexus Nats script.  When recalculating scores, we need to know this so as not to award denial points to a team-mate.
    HandicapStyle: char;      // default style of handicapping to use
    SortTeamsByRank,
    SoloHandicapsShowPercentage,
    AutoUpdateScoreboard,
    AutoUpdateTeams: Boolean;
    ScoreboardStyle: TScoreboardStyle;
    ElimMultiplier: Integer;  // If set, this is the number by which scores for players who have _not_ been eliminated should be multiplied by.
    StatusServerIP: string;   // IP address of the GameServer PC that we can ask for status packets
    DBServerIP: string;       // IP address of database server.
    AliasAutocomplete: Integer;  // In the AliasLaserforceDialog, how many characters does the user have to type before we autocomplete? 0 means 0 -- we pre-poulate the dialog with all possible members.
    GamesLimit: Integer;      // For Laserforce, how many games should we limit ourselves to when we query the system for the list of recent games?
    LeaguePoints: TDoubleList;  // victory points scored for first place, second place, etc.
    LeaguePointsHigh: Double;  // victory points scored for the high-scoring player
    LeaguePointsProportional: Double;  // victory points to be shared out to teams proportional to their game scores
  end;

function  StrToFloatDef(Val: string; Default: Real): Real;
function  StrToPercentDef(Val: string; Default: Integer): Integer;
procedure SendUDPs(port: Integer; s: string);
procedure SendUDP(port: Integer; s: string);
function  GetTable(address: string; port: Integer; s: string): TZoomReport;
function  GetHandiScore(SubItems: TStrings): Integer;
procedure SaveLeagueAsXml(FileName: string; var Prefs: TPrefs; var LeagueList: TLeagueList; var GameList: TGameList);
procedure LoadLeagueAsXml(FileName: string; var Prefs: TPrefs; var LeagueList: TLeagueList; var GameList: TGameList);
procedure ReadLeagueListAsXml(var F: TextFile; var LeagueList: TLeagueList);
procedure MakeBackup(FileName: string);
function  IsPoints(prefs: TPrefs) : Boolean;  // True if this league is scored on victory points, rather than just total scores.
function  NonNull(v: Variant): string;
procedure BubbleSort(SortList: TList; Compare: TListSortCompare);

var
  buffer: string;  // buffer always contains the next line of the XML file.

implementation

uses Windows, Dialogs, TornTeamBox1;

// convert a string to a float if it's valid, otherwise return a default value
function StrToFloatDef(Val: string; Default: Real): Real;
begin
  if (Val = '') or (Val[1] in ['a'..'z', 'A'..'Z']) then
    Result := Default
  else
    try
      Result := StrToFloat(Val);
    except
      on EConvertError do Result := Default;
    end;
end;

// convert a string to a float if it's valid, otherwise return a default value
function StrToPercentDef(Val: string; Default: Integer): Integer;
begin
  try
    if (Pos('%', Val) = 0) then
      Result := Round(StrToFloat(Val) * 10)
    else
      Result := Round(StrToFloat(Copy(Val, 1, Pos('%', Val) - 1)) * 10);
  except
    on EConvertError do Result := Default;
  end;
end;

function EncodeUrl(s: string): string;  // RFC 3986 section 2.3
var i: Integer;
begin
  Result := '';
  for i := 1 to Length(s) do
    if (s[i] in ['A'..'Z', 'a'..'z', '0'..'9', '-', '_', '.', '~']) then
      Result := Result + s[i]
    else
      Result := Result + '%' + IntToHex(Ord(s[i]), 2);
end;

procedure SendUDPs(port: Integer; s: string);
const CHUNK_LENGTH = 510;
var num: Integer;
begin
  num := 0;
  while Length(s) > CHUNK_LENGTH do begin  // loop until we've sent most of the string as chunks
    SendUDP(port, Format('%2.2d%s', [num, Copy(s, 1, CHUNK_LENGTH)]));
    s := Copy(s, CHUNK_LENGTH + 1, MaxInt);
    Inc(num);
  end;

  if Length(s) > 0 then begin  // send the leftover short chunk of the string
    SendUDP(port, Format('%2.2d%s', [num, s]));
    Inc(num);
  end;

  SendUDP(port, Format('%2.2d', [num]));  // send an empty chunk to terminate
end;

// Open a UDP socket on the listed port, broadcast a command, and close the socket.
//
// Print Server (port 12004) command format:
// PRINT,<printer number>,<printer number>,<sql query>,""
// Param1 = Print command (must be uppercase)
// Param2 = Print Form (this always matches the printer number)
// Param3 = Printer number
// Param4 = SQL Query
// Param5 = Arguments (always blank but must exist)
// e.g. PRINT,1,1,"Game=1 AND PackID IN (1,2)",""
//
// Scoreboard (port 12005) command format:
// PROCESSSCORES,<game number>,<command>
// <command> is ONCE (for end-of-game) or ALWAYS (for real-time)
// e.g. PROCESSSCORES,1,ONCE
//
procedure SendUDP(port: Integer; s: string);
var mysocket:   TSocket;    // handle to the Windows socket
    wsaData:    TWSAData;   // structure to receive details of the Windows socket, from WSAStartup
    sendbuf:    array[0..6000] of char;
    addr:       TSockAddrIn; // internet socket address structure
    one:        Integer;    // Integer variable containing the value 1.  Bloody Microsoft.
begin
  WSAStartup(MakeWord(1, 1), WSAData);        // request version 1.1 of the Windows Socket system
  mySocket := socket(AF_INET, SOCK_DGRAM, 0); // create an Internet Datagram (i.e. UDP) socket

  addr.sin_family := AF_INET;
  addr.sin_port   := htons(port);
//  addr.sin_addr.s_addr := htonl(INADDR_BROADCAST);
  addr.sin_addr.S_un_b.s_b1 := #255;  // These four lines do the same
  addr.sin_addr.S_un_b.s_b2 := #255;  // thing as the one line above, but
  addr.sin_addr.S_un_b.s_b3 := #255;  // the one line above produces
  addr.sin_addr.S_un_b.s_b4 := #255;  // a compiler warning.  Bah.

  one := 1;
  if setsockopt(mysocket, SOL_SOCKET, SO_BROADCAST, pchar(@one), sizeof(one)) = SOCKET_ERROR then  // Set socket option to enable broadcast.
    ShowMessage('Socket Option error ' + IntToStr(WSAGetLastError) + ' occurred.');

  StrPCopy(sendbuf, s);
  if sendto(mysocket, sendbuf, StrLen(sendbuf)+1, 0, addr, sizeof(addr)) = SOCKET_ERROR then  // broadcast message to UDP port
    ShowMessage('Socket Sendto error ' + IntToStr(WSAGetLastError) + ' occurred.');

  if closesocket(mysocket) = SOCKET_ERROR then
    ShowMessage('Closesocket error ' + IntToStr(WSAGetLastError) + ' occurred.');

  WSACleanup;
end;

function SendTCP(address: string; port: Integer; s: string): TSocket;
var wsaData:    TWSAData;   // structure to receive details of the Windows socket, from WSAStartup
    buf:    array[0..2000] of char;
    addr:       TSockAddrIn; // internet socket address structure
begin
  addr.sin_family := AF_INET;
  addr.sin_port   := htons(port);
  StrPCopy(buf, address);
  addr.sin_addr.s_addr := inet_addr(buf);

  WSAStartup(MakeWord(1, 1), WSAData);        // request version 1.1 of the Windows Socket system

  Result := socket(AF_INET, SOCK_STREAM, IPPROTO_TCP); // create an SOCK_STREAM (i.e. TCP) socket
  if Result = SOCKET_ERROR then  // send message
    ShowMessage('Socket creation error ' + IntToStr(WSAGetLastError) + ' occurred.');

  if connect(Result, addr, sizeof(addr)) = SOCKET_ERROR then  // send message
    ShowMessage('Socket Connect error ' + IntToStr(WSAGetLastError) + ' occurred.'#10'Check that Torn Connector is running.')
  else begin
    StrPCopy(buf, s);
    if sendto(Result, buf, StrLen(buf), 0, addr, sizeof(addr)) = SOCKET_ERROR then  // send message
        ShowMessage('Socket Sendto error ' + IntToStr(WSAGetLastError) + ' occurred.');
  end;
end;

procedure CloseTCP(socket: TSocket);
begin
  if closesocket(socket) = SOCKET_ERROR then
    ShowMessage('Closesocket error ' + IntToStr(WSAGetLastError) + ' occurred.');

  WSACleanup;
end;

function GetTCP(address: string; port: Integer; s: string): string;
var
    socket: TSocket;
    receivebuf: array[0..2000] of char;
    len: Integer;
begin
  socket := SendTCP(address, port, s + #13#10);
  try
    len := 1;
    while (len > 0) and (Pos('</body>', Result) = 0) do begin
      receivebuf[0] := #0;
      len := recv(socket, receivebuf, 1999, 0);  // Receive the packet and store it in receivebuf.
      Result := Result + Copy(receivebuf, 1, len);
    end;
  finally
    CloseTCP(socket);
  end;
end;

function GetTable(address: string; port: Integer; s: string): TZoomReport;
var
    response: string;
    section: string;
    row: string;
    attribute: string;
    cell: string;
    substart, subend: Integer;
    cellstart, cellend: Integer;
    zrow: TZRow;
begin
  response := GetTCP(address, port, s);//'GET ?' + EncodeUrl(s)+ ' HTTP/1.0'#10'From: Torn'#10'User-Agent: Torn/4.0'#10);
  Result := TZoomReport.Create;
  Result.Title := s;

  substart := Pos('<thead>', response) + 7;
  subend := Pos('</thead>', response);
  if (substart = 7) or (subend = 0) then Exit;
  row := Copy(response, substart, subend - substart);  // Everything between <thead> and </thead>.

  substart := Pos('<th class="', row) + 11;
  while (substart > 11) do begin
    subend := Pos('">', row);
    attribute := Copy(row, substart, subend - substart);  // Everything between <th class=" and ">.
    substart := subend + 2;  // 2 for <"; 11 for <th class="
    subend := Pos('</th>', row);
    cell := Copy(row, substart, subend - substart);  // Everything between "> and </th>.
    Result.Columns.Add(cell, '');
    if attribute = 'System.Int32' then
      Result.Columns[Result.Columns.Count - 1].ColumnType := ctInteger
    else
      Result.Columns[Result.Columns.Count - 1].ColumnType := ctString;

    row := Copy(row, subend + 1, MaxInt);
    substart := Pos('<th class="', row) + 11;
  end;

  substart := Pos('<tbody>', response) + 7;
  subend := Pos('</tbody>', response);
  if (substart = 0) or (subend = 0) then Exit;
  section := Copy(response, substart, subend - substart);  // Everything between <tbody> and </tbody>.

  substart := Pos('<tr>', section) + 4;
  subend := Pos('</tr>', section);
  while (substart > 4) and (subend >= substart) do begin
    row := Copy(section, substart, subend - substart);  // Everything between <tr> and </tr>.
    zrow := TZRow.Create;
    Result.Rows.Add(zrow);

    cellstart := Pos('<td>', row) + 4;
    cellend := Pos('</td>', row);
    while (cellstart > 4) and (cellend >= cellstart) do begin
      cell := Copy(row, cellstart, cellend - cellstart);  // Everything between <td> and </td>.
      zrow.Add(cell);
      row := Copy(row, cellend + 1, MaxInt);
      cellstart := Pos('<td>', row) + 4;
      cellend := Pos('</td>', row);
    end;

    section := Copy(section, subend + 5, MaxInt);
    substart := Pos('<tr>', section) + 4;
    subend := Pos('</tr>', section);
  end;
end;

// get a player's handicapped score out of a TeamBox SubItem
function GetHandiScore(SubItems: TStrings): Integer;
begin
  if Pos(' (', SubItems[tbScore]) = 0 then
    Result := StrToIntDef(SubItems[tbScore], 0)
  else
    Result := StrToIntDef(Copy(SubItems[tbScore],
                               Pos(' (', SubItems[tbScore]) + 2,
                               Pos(')',  SubItems[tbScore]) - Pos(' (', SubItems[tbScore]) - 2),  // yuck
                          0);
end;

procedure SaveLeagueAsXml(FileName: string; var Prefs: TPrefs; var LeagueList: TLeagueList; var GameList: TGameList);

  function EscapeXml(s: string): string;
  begin
    Result := StringReplace(StringReplace(StringReplace(s, '&', '&amp;', [rfReplaceAll]), '<', '&lt;', [rfReplaceAll]), '>', '&gt;', [rfReplaceAll]);
  end;

  procedure WriteBegin(var F: TextFile; var level: Integer; name: string);
  begin
    WriteLn(F, StringOfChar(' ', level * 2) + '<' + name + '>');
    Inc(level);
  end;

  procedure WriteEnd(var F: TextFile; var level: Integer; name: string);
  begin
    Dec(level);
    WriteLn(F, StringOfChar(' ', level * 2) + '</' + name + '>');
  end;

  procedure WriteValue(var F: TextFile; level: Integer; name, value: string); overload;
  begin
    if value <> '' then
      WriteLn(F, StringOfChar(' ', level * 2) + '<' + EscapeXml(name) + '>' + value + '</' + name + '>');
  end;

  procedure WriteValue(var F: TextFile; level: Integer; name: string; value: Integer); overload;
  begin
    WriteValue(F, level, name, IntToStr(value));
  end;

  procedure WriteValue(var F: TextFile; level: Integer; name: string; value: Double); overload;
  begin
    if value - Round(value) < 0.0000001 then
      WriteValue(F, level, name, IntToStr(Round(value)))
    else
      WriteValue(F, level, name, FloatToStr(value));
  end;

  procedure WriteValue(var F: TextFile; level: Integer; name: string; value: Boolean); overload;
  begin
    if value then
      WriteValue(F, level, name, '1')
    else
      WriteValue(F, level, name, '0');
  end;

var i, j: Integer;
    level: Integer;
    F: TextFile;
begin  //---SaveLeagueAsXml---
  try
    AssignFile(F, FileName);
    Rewrite(F);

    level := 0;
    try
      WriteLn(F, '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>');
      WriteBegin(F, level, 'body');

      WriteValue(F, level, 'GridHigh',         Prefs.GridHigh);
      WriteValue(F, level, 'GridWide',         Prefs.GridWide);
      WriteValue(F, level, 'GridPlayers',      Prefs.GridPlayers);
      WriteValue(F, level, 'SortMode',         Prefs.SortMode);
      WriteValue(F, level, 'HandicapStyle',    Prefs.HandicapStyle);
      WriteValue(F, level, 'SortByRank',       Prefs.SortTeamsByRank);
      WriteValue(F, level, 'AutoUpdate',       Prefs.AutoUpdateScoreboard);
      WriteValue(F, level, 'UpdateTeams',      Prefs.AutoUpdateTeams);
      WriteValue(F, level, 'ElimMultiplier',   Prefs.ElimMultiplier);

      for i := 0 to Prefs.LeaguePoints.Count - 1 do
        WriteValue(F, level, 'Points', Prefs.LeaguePoints[i]);
      WriteValue(F, level, 'High', Prefs.LeaguePointsHigh);
      WriteValue(F, level, 'Proportional', Prefs.LeaguePointsProportional);

      WriteBegin(F, level, 'leaguelist');
      for i := 0 to LeagueList.Count - 1 do
        with LeagueList[i] do begin
            WriteBegin(F, level, 'team index="' + IntToStr(i) + '"');
            WriteValue(F, level, 'teamname', TeamName);
            WriteValue(F, level, 'teamid',   TeamID);
            WriteValue(F, level, 'handicap', Handicap);
            WriteValue(F, level, 'comment',  Comment);

            WriteBegin(F, level, 'players');
            for j := 0 to Players.Count - 1 do
              with Players[j] do begin
                WriteBegin(F, level, 'player');
                WriteValue(F, level, 'name',     Name);
                WriteValue(F, level, 'buttonid', ButtonID);
                WriteValue(F, level, 'handicap', Handicap);
                WriteValue(F, level, 'comment',  Comment);
                WriteEnd(F, level, 'player');
              end;
            WriteEnd(F, level, 'players');
            WriteEnd(F, level, 'team');
        end;
      WriteEnd(F, level, 'leaguelist');

      WriteBegin(F, level, 'games');
      for i := 0 to GameList.Count - 1 do begin
        with GameList[i] do begin
          WriteBegin(F, level, 'game index="' + IntToStr(i) + '"');
          if Title <> '' then WriteValue(F, level, 'title', Title);
          WriteValue(F, level, 'gametime', FloatToStr(GameTime));
          WriteValue(F, level, 'ansigametime', FormatDateTime('yyyy/mm/dd hh:mm:ss', GameTime));
          WriteValue(F, level, 'hits', IntToStr(GameList[i].Hits));
          if Secret then
            WriteValue(F, level, 'secret', 'y');

          WriteBegin(F, level, 'teams');
          for j := 0 to Teams.Count - 1 do
            with Teams[j] do begin
              WriteBegin(F, level, 'team index="' + IntToStr(j) + '"');
              WriteValue(F, level, 'teamid', TeamID);
              if (Colour = '') and (NexusColour <> -1) then
                WriteValue(F, level, 'colour', NexusColourNames[NexusColour])
              else
                WriteValue(F, level, 'colour', Colour);
              WriteValue(F, level, 'score',  Score);
              WriteValue(F, level, 'points', Points);
              WriteValue(F, level, 'adjustment', Adjustment);
              WriteValue(F, level, 'victorypointsadjustment', PointsAdjustment);
              WriteEnd  (F, level, 'team');
            end;
          WriteEnd(F, level, 'teams');

          WriteBegin(F, level, 'players');
          for j := 0 to Players.Count - 1 do
            with Players[j] do begin
              WriteBegin(F, level, 'player index="' + IntToStr(j) + '"');
              WriteValue(F, level, 'teamid',   TeamID);
              WriteValue(F, level, 'playerid', PlayerID);
              WriteValue(F, level, 'pack',     Pack);
              WriteValue(F, level, 'score',    Score);
              WriteValue(F, level, 'rank',     Rank);
              WriteValue(F, level, 'hitsby',   HitsBy);
              WriteValue(F, level, 'hitson',   HitsOn);
              if BaseHits <> 0 then
                WriteValue(F, level, 'basehits', BaseHits);
              if BaseDestroys <> 0 then
                WriteValue(F, level, 'basedestroys', BaseDestroys);
              WriteValue(F, level, 'colour',   Colour);
              WriteEnd(F, level, 'player');
            end;
          WriteEnd(F, level, 'players');
        end;
        WriteEnd(F, level, 'game');
      end;  // for i to GameList
      WriteEnd(F, level, 'games');
      WriteEnd(F, level, 'body');
    finally
      CloseFile(F);
    end;
  except
    on E: EInOutError do begin
      MessageDlg('Error "' + E.Message + '" occurred while opening file "' + FileName +
        '".'#13#10'Check that the file is accessible; for example if it''s on the network, make sure that the network location is reachable.', mtError, [mbOK], 0);
      raise;  // Gotta re-raise so that the caller knows not to flag the league file as successfully opened.
    end;
  end;
end;

  function ReadBegin(var F: TextFile; name: string): Boolean;
  begin
      Result := (Pos('<' + name, buffer) > 0) and (Pos('>', buffer) > 0) and (Pos('<' + name + '/>', buffer) = 0) and (Pos('<' + name + ' />', buffer) = 0);  // True if tag found but is not self-closing (like <br/>).
      if (Pos('<' + name, buffer) > 0) and
         ((Pos('>', buffer) > 0) or (Pos('<' + name + '/>', buffer) = 0) or (Pos('<' + name + ' />', buffer) = 0)) then
        ReadLn(F, buffer);  // buffer did contain the begin tag we were looking for (self-closing or not). Clear that begin tag out of the buffer and move forward.
//      else
//        Exceptions := Exceptions + ',"' + DateTimeToStr(Now) + ': ReadBegin: expected <' + name + '>, but found ' + Trim(buffer) + '."'
  end;

  procedure ReadEnd(var F: TextFile; name: string);
  begin
//      if Pos('</' + name + '>', buffer) = 0 then
//        Exceptions := Exceptions + ',"' + DateTimeToStr(Now) + ': ReadEnd: expected <' + name + '>, but found ' + Trim(buffer) + '."';

      ReadLn(F, buffer);
  end;

  function ReadRawString(var F: TextFile; name: string; Default: string = ''): string;
  begin
    if Pos('<' + name + '/>', buffer) > 0 then begin
      Result := Default;
      ReadLn(F, buffer);
    end else if Pos('<' + name + '>', buffer) = 0 then begin
      Result := Default;
//            Exceptions := Exceptions + ',"' + DateTimeToStr(Now) + ': ReadString: expected <' + name + '>, but found ' + Trim(buffer) + '."'
    end else begin
      buffer := Copy(buffer, Pos('<' + name + '>', buffer) + Length(name) + 2, MaxInt);
      Result := Copy(buffer, 1, Pos('</', buffer) - 1);
      ReadLn(F, buffer);
    end;
  end;

  function ReadString(var F: TextFile; name: string; Default: string = ''): string;
  begin
    Result := StringReplace(StringReplace(StringReplace(ReadRawString(F, name, Default), '&gt;', '>', [rfReplaceAll]), '&lt;', '<', [rfReplaceAll]), '&amp;', '&', [rfReplaceAll]);
  end;

  function ReadInteger(var F: TextFile; name: string; Default: Integer = 0): Integer;
  begin
    Result := StrToIntDef(ReadRawString(F, name), Default);
  end;

  function ReadFloat(var F: TextFile; name: string; Default: Double = 0): Double;
  begin
    Result := StrToFloatDef(ReadRawString(F, name), Default);
  end;

  function ReadBool(var F: TextFile; name: string; Default: Boolean): Boolean;
  begin
    if Default then
      Result := ReadRawString(F, name, '1') <> '0'
    else
      Result := ReadRawString(F, name, '0') <> '0';
  end;

  // Return the tag name, stripped of angle brackets but not any leading or trailing slash.
  function Peek: string; overload;
  begin
    if (Pos('<', buffer) > 0) and (Pos('>', buffer) > 0) then
      Result := Copy(buffer, Pos('<', buffer) + 1,
                     Pos('>', buffer) - Pos('<', buffer) - 1)
    else
      Result := '';
  end;

  function Peek(name: string): Boolean; overload;
  begin
    Result := Pos(name, Peek) = 1;
  end;

  procedure Discard(var F: TextFile);
  var s: string;
  begin
    s := Peek;
    ReadString(F, s);  // discard unknown tag
//        Exceptions := Exceptions + ',"' + DateTimeToStr(Now) + ': Found unknown tag ' + s + '."';
  end;

  function DiscardIfEmpty(var F: TextFile; name: string): Boolean; overload;
  var
    i: Integer;
    s: string;
  begin
    Result := False;
    s := Peek;
    i := Pos(name, s);
    if i = 1 then begin
      Inc(i, Length(name));
      while (i <= Length(s)) and (s[i] = ' ') do
        Inc(i);
      Result := (i <= Length(s)) and (s[i] = '/');
      if Result then
        ReadString(F, s);  // discard line
    end;
  end;


procedure LoadLeagueAsXml(FileName: string; var Prefs: TPrefs; var LeagueList: TLeagueList; var GameList: TGameList);
var
  F: TextFile;
  s: string;
  hasgametime: Boolean;

  function Parse(s: string): TDateTime;
  var year, month, day, hour, minute, second: Integer;
  begin  // yyyy-mm-dd hh:mm:ss
    try
      year := StrToInt(Copy(s, 1, 4));
      month := StrToInt(Copy(s, 6, 2));
      day := StrToInt(Copy(s, 9, 2));
      hour := StrToInt(Copy(s, 12, 2));
      minute := StrToInt(Copy(s, 15, 2));
      second := StrToInt(Copy(s, 18, 2));
      Result := EncodeDate(year, month, day) + EncodeTime(hour, minute, second, 0);
    except
      Result := 0;
    end;
  end;

begin
  LeagueList.Clear;
  GameList.Clear;

  try
    AssignFile(F, FileName);
    Reset(F);

    try
      ReadLn(F);  // skip the xml version line
      ReadLn(F, buffer);  // load the buffer
      ReadBegin(F, 'body');

      while not Eof(F) and not Peek('Points')and not Peek('leaguelist') do  // read a whole bunch of tags in no particular order, and which may or may not be present
             if Peek('GridHigh') then         Prefs.GridHigh          := ReadInteger(F, 'GridHigh',          Prefs.GridHigh)
        else if Peek('GridWide') then         Prefs.GridWide          := ReadInteger(F, 'GridWide',          Prefs.GridWide)
        else if Peek('GridPlayers') then      Prefs.GridPlayers       := ReadInteger(F, 'GridPlayers',       Prefs.GridPlayers)
        else if Peek('SortMode') then         Prefs.SortMode          := ReadInteger(F, 'SortMode',          Prefs.SortMode)
        else if Peek('HandicapStyle') then    Prefs.HandicapStyle     :=  ReadString(F, 'HandicapStyle',     Prefs.HandicapStyle)[1]  // the [1] at the end is to convert string to char.
        else if Peek('SortByRank') then       Prefs.SortTeamsByRank   :=    ReadBool(F, 'SortByRank',        Prefs.SortTeamsByRank)
        else if Peek('AutoUpdate') then       Prefs.AutoUpdateScoreboard := ReadBool(F, 'AutoUpdate',        Prefs.AutoUpdateScoreboard)
        else if Peek('UpdateTeams') then      Prefs.AutoUpdateTeams   :=    ReadBool(F, 'UpdateTeams',       Prefs.AutoUpdateTeams)
        else if Peek('ElimMultiplier') then   Prefs.ElimMultiplier   :=  ReadInteger(F, 'ElimMultiplier',    Prefs.ElimMultiplier)
        else Discard(F);

      Prefs.LeaguePoints := TDoubleList.Create;
      while not Eof(F) and Peek('Points') do
        Prefs.LeaguePoints.Add(ReadFloat(F, 'Points', 0));
      if Peek('High') then Prefs.LeaguePointsHigh := ReadInteger(F, 'High', 0);
      if Peek('Proportional') then Prefs.LeaguePointsProportional := ReadInteger(F, 'Proportional', 0);

      ReadLeagueListAsXml(F, LeagueList);

      if not DiscardIfEmpty(F, 'games') then begin
        ReadBegin(F, 'games');
        while not Eof(F) and not Peek('/games') do begin
          with GameList.Add(TGame.Create) do begin
            ReadBegin(F, 'game');
            if Peek('title') then Title := ReadString(F, 'title');
            hasgametime := Peek('gametime');
            if hasgametime then GameTime := StrToFloat(ReadString(F, 'gametime'));
            if Peek('ansigametime') then begin
              s := ReadString(F, 'ansigametime');
              if not hasgametime then
                GameTime := Parse(s);
            end;
            Hits := ReadInteger(F, 'hits');
            if Peek('secret') then begin
              s := ReadString(F, 'secret');
              Secret := s = 'y';
            end;
            ReadBegin(F, 'teams');
            while not Eof(F) and not Peek('/teams') do
              with Teams.Add do begin
                ReadBegin(F, 'team');
                while not Eof(F) and not Peek('/team') do
                       if Peek('teamid') then TeamID := ReadInteger(F, 'teamid')
                  else if Peek('colour') then Colour := ReadString(F,  'colour')
                  else if Peek('score') then  Score  := ReadInteger(F, 'score')
                  else if Peek('points') then Points := ReadInteger(F, 'points')
                  else if Peek('adjustment') then Adjustment := ReadInteger(F, 'adjustment')
                  else if Peek('victorypointsadjustment') then PointsAdjustment := ReadInteger(F, 'victorypointsadjustment')
                  else Discard(F);

                ReadEnd(F, 'team');
              end;
            ReadEnd(F, 'teams');

            if ReadBegin(F, 'players') then
              while not Eof(F) and not Peek('/players') do
                with Players.Add do begin
                  ReadBegin(F, 'player');
                  while not Eof(F) and not Peek('/player') do
                       if Peek('teamid') then   TeamID   := ReadInteger(F, 'teamid')
                  else if Peek('playerid') then PlayerID := ReadString(F,  'playerid')
                  else if Peek('pack') then     Pack     := ReadString(F,  'pack')
                  else if Peek('score') then    Score    := ReadInteger(F, 'score')
                  else if Peek('rank') then     Rank     := ReadInteger(F, 'rank')
                  else if Peek('hitsby') then   HitsBy   := ReadInteger(F, 'hitsby')
                  else if Peek('hitson') then   HitsOn   := ReadInteger(F, 'hitson')
                  else if Peek('basehits') then BaseHits := ReadInteger(F, 'basehits')
                  else if Peek('basedestroys') then BaseDestroys := ReadInteger(F, 'basedestroys')
                  else if Peek('colour') then   Colour   := ReadInteger(F, 'colour')
                  else Discard(F);

                  ReadEnd(F, 'player');
                end;
            ReadEnd(F, 'players');
          end;  // with Game
          ReadEnd(F, 'game');
        end;  // while not Peek('/games')
        ReadEnd(F, 'games');
      end;
      ReadEnd(F, 'body');
    finally
      CloseFile(F);
    end;
  except
    on E: EInOutError do begin
      MessageDlg('Error "' + E.Message + '" occurred while opening file "' + FileName +
        '".'#13#10'Check that the file is accessible; for example if it''s on the network, make sure that the network location is reachable.', mtError, [mbOK], 0);
      raise;  // Gotta re-raise so that the caller knows not to flag the league file as successfully opened.
    end;
  end;
end;

procedure ReadLeagueListAsXml(var F: TextFile; var LeagueList: TLeagueList);
begin
  if buffer = '' then ReadLn(F, buffer);  // load the buffer

  while not Eof(F) and not Peek('leaguelist') do Discard(F);

  ReadBegin(F, 'leaguelist');
  while not Eof(F) and not Peek('/leaguelist') do
    with LeagueList.Add do begin
      ReadBegin(F, 'team');
      while not Eof(F) and not Peek('players') and not Peek('/team') do
             if Peek('teamname') then TeamName := ReadString(F,  'teamname')
        else if Peek('teamid') then   TeamID   := ReadInteger(F, 'teamid')
        else if Peek('handicap') then Handicap := ReadString(F,  'handicap')
        else if Peek('comment') then  Comment  := ReadString(F,  'comment')
        else Discard(F);

      if Peek('/team') then ReadEnd  (F, 'team');

      if not DiscardIfEmpty(F, 'players') then begin
        ReadBegin(F, 'players');
        while not Peek('/players') and not (buffer = '') do
          with Players.Add do begin
            ReadBegin(F, 'player');
            while not Eof(F) and not Peek('/player') do
                   if Peek('name') then     Name     := ReadString(F, 'name')
              else if Peek('buttonid') then ButtonID := ReadString(F, 'buttonid')
              else if Peek('handicap') then Handicap := ReadString(F, 'handicap')
              else if Peek('comment') then  Comment  := ReadString(F, 'comment')
              else Discard(F);

            ReadEnd(F, 'player');
          end;
        ReadEnd(F, 'players');
      end;
      if Peek('/team') then ReadEnd  (F, 'team');
    end;
  ReadEnd(F, 'leaguelist');
end;

// Erase any existing .TornBackup file.  Rename the passed file as .TornBackup.
// But only do these things if the new file is bigger than the old file.
procedure MakeBackup(FileName: string);
var F: TextFile;
    backupname: string;
    oldsize, newsize: Integer;
begin
  try
    backupname := ChangeFileExt(FileName, '.TornBackup');
    if FileExists(backupname) then begin
      AssignFile(F, backupname);
      Reset(F);
      oldsize := FileSize(F);  // get the size of the backup
      CloseFile(F);
    end else
      oldsize := -1;

    AssignFile(F, FileName);
    Reset(F);
    newsize := FileSize(F);  // get the size of the new backup-to-be
    CloseFile(F);
    if newsize > oldsize then begin
      if FileExists(backupname) then
        SysUtils.DeleteFile(backupname);  // delete the old backup

      AssignFile(F, FileName);
      Rename(F, backupname);  // keep the .Torn file as the new backup
    end;
  finally
    // squish
  end;
end;

function IsPoints(prefs: TPrefs) : Boolean;
var i: Integer;
begin
  Result := (prefs.LeaguePointsHigh > 0) or (prefs.LeaguePointsProportional > 0);

  if not Result then
    for i := 0 to prefs.LeaguePoints.Count - 1 do
      if prefs.LeaguePoints[i] <> 0 then begin
        Result := True;
        Break;
      end;
end;

function NonNull(v: Variant): string;
begin
  if VarIsNull(v) then
    Result := ''
  else
    Result := v;
end;

// Bubblesort a list, using the given Compare function.
// Partly nicked from Classes.pas TList.Sort and QuickSort, to work around a bug in Delphi 5's QuickSort.
procedure BubbleSort(SortList: TList; Compare: TListSortCompare);
var I, J: Integer;
begin
    if (SortList = nil) or (SortList.Count < 2) then Exit;
    for I := 0 to SortList.Count - 1 do
        for J := I to SortList.Count - 1 do
            if Compare(SortList[I], SortList[J]) > 0 then
                SortList.Exchange(I, J);
end;

end.
